

def compare_words_tokens(tokens, words, word_ids):
    previous_word_id = None
    new_words = []
    for index, word_id in enumerate(word_ids):
        if word_id is not None and previous_word_id == word_id:
            new_words[-1] += tokens[index][2:]
        elif word_id is not None and previous_word_id != word_id:
            new_words.append(tokens[index])
        previous_word_id = word_id

    print(words)
    print(new_words)
    for index, new_word in enumerate(new_words):
        if words[index].lower() != new_word:
            return False
    else:
        return True



def add_pad_for_labels(word_ids, labels, pad_id, token_type_ids, label_all_tokens=False, required_token_type_id=0, return_criterion_mask=False):

    new_labels = []
    criterion_mask = []
    previous_word_id = None
    for index, word_id in enumerate(word_ids):
        if word_id is None or token_type_ids[index] != required_token_type_id:
            new_labels.append(pad_id)
            criterion_mask.append(0)
        elif word_id !=  previous_word_id:
            new_labels.append(labels[word_id])
            criterion_mask.append(1)
        else:
            new_labels.append(labels[-1] if label_all_tokens is True else pad_id)
            criterion_mask.append(0)
        previous_word_id = word_id

    if return_criterion_mask is False:
        return new_labels
    return new_labels, criterion_mask



def remove_pad_for_labels(word_ids, labels, token_type_ids, required_token_type_id=0):
    new_labels = []

    previous_word_id = None
    for index, word_id in enumerate(word_ids):
        if word_id is None or token_type_ids[index] != required_token_type_id or word_id ==  previous_word_id:
            continue
        else:
            new_labels.append(labels[index])
        previous_word_id = word_id
    return new_labels


def add_pad_for_2d_labels(word_ids, labels, pad_id, token_type_ids, label_all_tokens=False, required_token_type_id=0, return_criterion_mask=False):

    new_labels = [[pad_id for _ in word_ids] for _ in word_ids]
    criterion_mask = [[0 for _ in word_ids] for _ in word_ids]
    x_previous_word_id = None
    for x_index, x_word_id in enumerate(word_ids):
        if token_type_ids[x_index] != required_token_type_id or x_word_id is None or (label_all_tokens is False and x_previous_word_id == x_word_id):
            continue

        y_previous_word_id = None
        for y_index, y_word_id in enumerate(word_ids):
            if token_type_ids[y_index] != required_token_type_id or y_word_id is None or (label_all_tokens is False and y_previous_word_id == y_word_id):
                continue

            if y_previous_word_id != y_word_id:
                criterion_mask[x_index][y_index] = 1
            new_labels[x_index][y_index] = labels[x_word_id][y_word_id]
            y_previous_word_id = y_word_id

        x_previous_word_id = x_word_id

    if return_criterion_mask is False:
        return new_labels
    return new_labels, criterion_mask


def remove_pad_for_2d_labels(word_ids, labels, token_type_ids, required_token_type_id=0):

    new_labels = []

    x_previous_word_id = None
    for x_index, x_word_id in enumerate(word_ids):
        if token_type_ids[x_index] != required_token_type_id or x_word_id is None or x_previous_word_id == x_word_id:
            continue

        _new_labels = []

        y_previous_word_id = None
        for y_index, y_word_id in enumerate(word_ids):
            if token_type_ids[y_index] != required_token_type_id or y_word_id is None or y_previous_word_id == y_word_id:
                continue

            _new_labels.append(labels[x_index][y_index])
            y_previous_word_id = y_word_id

        new_labels.append(_new_labels)
        x_previous_word_id = x_word_id

    return new_labels


def add_pad_for_3d_labels(
        word_ids, 
        labels, 
        pad_id, 
        token_type_ids, 
        label_all_tokens=False, 
        required_token_type_id=0, 
        return_criterion_mask=False):
    new_labels = []
    criterion_mask = []
    for _labels in labels:
        _new_labels, _criterion_mask = add_pad_for_2d_labels(
                word_ids, 
                _labels, 
                pad_id, 
                token_type_ids, 
                label_all_tokens=label_all_tokens, 
                required_token_type_id=required_token_type_id, 
                return_criterion_mask=return_criterion_mask)
        new_labels.append(_new_labels)
        criterion_mask.append(_criterion_mask)

    if return_criterion_mask is False:
        return new_labels
    return new_labels, criterion_mask


def remove_pad_for_3d_labels(word_ids, labels, token_type_ids, required_token_type_id=0):

    return [
        remove_pad_for_2d_labels(word_ids, _labels, token_type_ids, required_token_type_id=required_token_type_id)
        for _labels in labels
    ]

    # for _labels in labels:
    #     import ipdb; ipdb.set_trace()
    #     _new_labels = remove_pad_for_2d_labels(word_ids, _labels, token_type_ids, required_token_type_id=required_token_type_id)
    #     print(_new_labels)




import torch
from dataclasses import dataclass
from torch.utils.data import Dataset
from typing import Dict, Optional, List
from . import NerExample, NerFeature, EntityExample
from ..utils import add_pad_for_3d_labels, remove_pad_for_3d_labels, compare_words_tokens


@dataclass
class GlobalPointerForNerFeature(NerFeature):
    labels: torch.Tensor
    criterion_mask: torch.Tensor


def convert_examples_to_feature(
        tokenizer,
        example_list: List[NerExample],
        label_to_id_map: Dict,
        max_seq_length: int,
        pad_id: int=-100,
        label_all_tokens: bool=False,
        check_tokenization=False
        ) -> List[GlobalPointerForNerFeature]:

    
    encoded_inputs = tokenizer(
        [example.words for example in example_list],
        is_split_into_words=True,
        truncation=True,
        padding='max_length',
        max_length=max_seq_length
    )

    feature_list = []
    for index, example in enumerate(example_list):
        word_ids = encoded_inputs.word_ids(index)
        assert check_tokenization is False or compare_words_tokens(tokenizer.convert_ids_to_tokens(encoded_inputs['input_ids'][index]), example.words, word_ids), example

        len_words = len(example.words)

        labels = [
            [
                [ 0 for _ in example.words]
                for _ in example.words
            ]
            for _ in label_to_id_map
        ]

        # read labels
        for entity in example.entities:
            if len_words < entity.end_idx or ''.join(example.words[entity.start_idx: entity.end_idx]) != entity.entity:
                guess_start = example.text.index(entity.entity)
                guess_end = guess_start + len(entity.entity)
                raise Exception(f'index error, text: {example.text}, item: {entity}, max_len: {len_words}, guess: {(guess_start, guess_end)}')
            labels[label_to_id_map[entity.type]][entity.start_idx][entity.end_idx-1] = 1


        # fix labels
        labels, criterion_mask = add_pad_for_3d_labels(
                word_ids, 
                labels, 
                pad_id, 
                encoded_inputs['token_type_ids'][index], 
                label_all_tokens, 
                return_criterion_mask=True)


        feature_list.append(GlobalPointerForNerFeature(
            text=example.text,
            entities=example.entities,
            words=example.words,
            word_ids=word_ids,
            input_ids=torch.tensor(encoded_inputs['input_ids'][index]),
            attention_mask=torch.tensor(encoded_inputs['attention_mask'][index]),
            token_type_ids=torch.tensor(encoded_inputs['token_type_ids'][index]),
            labels=torch.tensor(labels),
            criterion_mask=torch.tensor(criterion_mask)
        ))

    return feature_list



class GlobalPointerForNerDataset(Dataset):
    def __init__(self, example_list: List[NerExample], tokenizer, label_to_id_map, max_seq_length, pad_id=-100, label_all_tokens=False, check_tokenization=False, lazy_load=False):
        self.example_list = example_list
        self.tokenizer = tokenizer
        self.label_to_id_map = label_to_id_map
        self.max_seq_length = max_seq_length
        self.pad_id = pad_id
        self.label_all_tokens = label_all_tokens
        self.check_tokenization = check_tokenization
        self.lazy_load = lazy_load
        

        if lazy_load is False:
            self.feature_list = convert_examples_to_feature(
                    self.tokenizer, 
                    self.example_list, 
                    self.label_to_id_map, 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)

    def __len__(self):
        # return len(self.feature_list)
        return len(self.example_list)

    def __getitem__(self, idx) -> GlobalPointerForNerFeature:
        if self.lazy_load is False:
            return self.feature_list[idx]
        else:
            return convert_examples_to_feature(
                    self.tokenizer, 
                    [self.example_list[idx]], 
                    self.label_to_id_map, 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)[0]


def convert_logits_to_examples(
        feature_list: List[GlobalPointerForNerFeature], 
        logits: torch.Tensor, 
        id_to_label_map: Dict, 
        remove_overlap=None) -> List[NerExample]:
    """
    remove_overlap去重， 可取值:
        None: 不去重
        'short': 去除短的
        'long': 去除长的
    """

    example_list = []

    for logit, feature in zip(logits, feature_list):

        labels = torch.where(logit>0, 1, 0).tolist()
        labels = remove_pad_for_3d_labels(feature.word_ids, labels, feature.token_type_ids)

        entities = []
        for label_id, start_end_labels in enumerate(labels):
            previous_end_index = None
            for start_index, end_labels in enumerate(start_end_labels):
                for end_index, span_label in enumerate(end_labels):

                    if end_index >= start_index and span_label == 1:
                        if  remove_overlap == 'short' and end_index == previous_end_index:
                            continue
                        elif  remove_overlap == 'long' and end_index == previous_end_index:
                            entities.pop()

                        entities.append(EntityExample(
                            start_idx=start_index,
                            end_idx=end_index+1,
                            entity=''.join(feature.words[start_index: end_index+1]),
                            type=id_to_label_map[label_id]))
                        previous_end_index = end_index
                        break
        
        example_list.append(NerExample(
            text=feature.text, 
            entities=sorted(entities, key=lambda entity: entity.start_idx), 
            words=feature.words))

    return example_list



import torch
from torch import nn
from torch.nn import functional as F
from transformers import AutoModel 
from siumaai.losses.focal_loss import FocalLoss
from siumaai.losses.label_smoothing import LabelSmoothingCrossEntropy
from torch.nn.modules import BCEWithLogitsLoss

class MultiNonLinearClassifier(nn.Module):
    def __init__(self, hidden_size, num_label, dropout_rate):
        super(MultiNonLinearClassifier, self).__init__()
        self.num_label = num_label
        self.classifier1 = nn.Linear(hidden_size, hidden_size)
        self.classifier2 = nn.Linear(hidden_size, num_label)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, input_features):
        features_output1 = self.classifier1(input_features)
        # features_output1 = F.relu(features_output1)
        features_output1 = F.gelu(features_output1)
        features_output1 = self.dropout(features_output1)
        features_output2 = self.classifier2(features_output1)
        return features_output2


class MRCForNer(nn.Module):
    def __init__(self, pretrain_model_path, dropout_rate, hidden_size, vocab_len=None):
        super(MRCForNer, self).__init__()

        self.model = AutoModel.from_pretrained(pretrain_model_path)

        self.start_outputs = nn.Linear(hidden_size, 1)
        self.end_outputs = nn.Linear(hidden_size, 1)
        self.span_embedding = MultiNonLinearClassifier(hidden_size * 2, 1, dropout_rate)
        self.criterion = BCEWithLogitsLoss()

        if vocab_len is not None:
            self.model.resize_token_embeddings(vocab_len)

    def forward(
            self, 
            input_ids, 
            token_type_ids, 
            attention_mask, 
            start_labels=None, 
            end_labels=None, 
            span_labels=None, 
            start_criterion_mask=None,
            end_criterion_mask=None,
            span_criterion_mask=None):

        # [batch_size, seq_len, hidden_size]
        sequence_output =self.model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)[0]

        batch_size, seq_len, hid_size = sequence_output.size()

        # [batch_size, seq_len]
        start_logits = self.start_outputs(sequence_output).squeeze(-1)
        end_logits = self.end_outputs(sequence_output).squeeze(-1)

        # [batch_size, seq_len, seq_len, hidden_size]
        start_extend = sequence_output.unsqueeze(2).expand(-1, -1, seq_len, -1)
        # [batch_size, seq_len, seq_len, hidden_size]
        end_extend = sequence_output.unsqueeze(1).expand(-1, seq_len, -1, -1)
        # [batch, seq_len, seq_len, hidden*2]
        span_matrix = torch.cat([start_extend, end_extend], 3)
        # [batch, seq_len, seq_len]
        span_logits = self.span_embedding(span_matrix).squeeze(-1)
        
        outputs = (start_logits, end_logits, span_logits)

        # if start_labels is not None and end_labels is not None and span_labels is not None:
        if None not in [start_labels, end_labels, span_labels, start_criterion_mask, end_criterion_mask, span_criterion_mask]:

            # start_logits_mask = attention_mask.view(-1) == 1
            # end_logits_mask = attention_mask.view(-1) == 1

            # x_mask = attention_mask.unsqueeze(-2).expand(-1, seq_len, -1)
            # y_mask = attention_mask.unsqueeze(-1).expand(-1, -1, seq_len)
            # span_logits_mask = torch.triu(x_mask & y_mask, 0).view(-1) == 1

            start_criterion_mask = start_criterion_mask.view(-1).bool()
            end_criterion_mask = end_criterion_mask.view(-1).bool()
            span_criterion_mask = torch.triu(span_criterion_mask, 0).view(-1).bool()

            active_start_logits = start_logits.view(-1)[start_criterion_mask]
            active_end_logits = end_logits.view(-1)[end_criterion_mask]
            active_span_logits = span_logits.view(-1)[span_criterion_mask]

            active_start_labels = start_labels.view(-1)[start_criterion_mask]
            active_end_labels = end_labels.view(-1)[end_criterion_mask]
            active_span_labels = span_labels.view(-1)[span_criterion_mask]

            start_loss = self.criterion(active_start_logits, active_start_labels)
            end_loss = self.criterion(active_end_logits, active_end_labels)
            span_loss = self.criterion(active_span_logits, active_span_labels)
            outputs = (start_loss+end_loss+span_loss, ) + outputs

        return outputs




import torch
from dataclasses import dataclass
from torch.utils.data import Dataset
from typing import Dict, Optional, List
from . import NerExample, NerFeature, EntityExample
from ..utils import add_pad_for_labels, remove_pad_for_labels, compare_words_tokens


@dataclass
class SpanForNerFeature(NerFeature):
    start_labels: Optional[torch.Tensor]=None
    end_labels: Optional[torch.Tensor]=None


def convert_examples_to_feature(
        tokenizer,
        example_list: List[NerExample],
        label_to_id_map: Dict,
        max_seq_length: int,
        pad_id: int=-100,
        label_all_tokens: bool=False,
        check_tokenization=False
        ) -> List[SpanForNerFeature]:

    
    encoded_inputs = tokenizer(
        [example.words for example in example_list],
        is_split_into_words=True,
        truncation=True,
        padding='max_length',
        max_length=max_seq_length
    )

    feature_list = []
    for index, example in enumerate(example_list):
        word_ids = encoded_inputs.word_ids(index)
        assert check_tokenization is False or compare_words_tokens(tokenizer.convert_ids_to_tokens(encoded_inputs['input_ids'][index]), example.words, word_ids), example

        len_words = len(example.words)
        start_labels = [label_to_id_map['O'] for _ in range(len_words)]
        end_labels = [label_to_id_map['O'] for _ in range(len_words)]

        # read labels
        for entity in example.entities:
            if len_words < entity.end_idx or ''.join(example.words[entity.start_idx: entity.end_idx]) != entity.entity:
                guess_start = example.text.index(entity.entity)
                guess_end = guess_start + len(entity.entity)
                raise Exception(f'index error, text: {example.text}, item: {entity}, max_len: {len_words}, guess: {(guess_start, guess_end)}')

            start_labels[entity.start_idx] = label_to_id_map[entity.type]
            end_labels[entity.end_idx-1] = label_to_id_map[entity.type]

        # fix labels
        start_labels = add_pad_for_labels(word_ids, start_labels, pad_id, encoded_inputs['token_type_ids'][index], label_all_tokens)
        end_labels = add_pad_for_labels(word_ids, end_labels, pad_id, encoded_inputs['token_type_ids'][index], label_all_tokens)


        feature_list.append(SpanForNerFeature(
            text=example.text,
            entities=example.entities,
            words=example.words,
            word_ids=word_ids,
            input_ids=torch.tensor(encoded_inputs['input_ids'][index]),
            attention_mask=torch.tensor(encoded_inputs['attention_mask'][index]),
            token_type_ids=torch.tensor(encoded_inputs['token_type_ids'][index]),
            start_labels=torch.tensor(start_labels),
            end_labels=torch.tensor(end_labels)
        ))

    return feature_list



class SpanForNerDataset(Dataset):
    def __init__(self, example_list: List[NerExample], tokenizer, label_to_id_map, max_seq_length, pad_id=-100, label_all_tokens=False, check_tokenization=False, lazy_load=False):
        self.example_list = example_list
        self.tokenizer = tokenizer
        self.label_to_id_map = label_to_id_map
        self.max_seq_length = max_seq_length
        self.pad_id = pad_id
        self.label_all_tokens = label_all_tokens
        self.check_tokenization = check_tokenization
        self.lazy_load = lazy_load
        

        if lazy_load is False:
            self.feature_list = convert_examples_to_feature(
                    self.tokenizer, 
                    self.example_list, 
                    self.label_to_id_map, 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)

    def __len__(self):
        # return len(self.feature_list)
        return len(self.example_list)

    def __getitem__(self, idx) -> SpanForNerFeature:
        if self.lazy_load is False:
            return self.feature_list[idx]
        else:
            return convert_examples_to_feature(
                    self.tokenizer, 
                    [self.example_list[idx]], 
                    self.label_to_id_map, 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)[0]


def convert_logits_to_examples(feature_list: List[SpanForNerFeature], start_logits: torch.Tensor, end_logits: torch.Tensor, id_to_label_map: Dict, remove_overlap=None) -> List[NerExample]:
    """
    remove_overlap去重， 可取值:
        None: 不去重
        'short': 去除短的
        'long': 去除长的
    """

    example_list = []
    for start_logit, end_logit, feature in zip(start_logits, end_logits, feature_list):
        start_label_id_list = start_logit.argmax(axis=1).tolist()
        start_label_id_list = remove_pad_for_labels(feature.word_ids, start_label_id_list, feature.token_type_ids)
        start_label_list = [id_to_label_map[label_id] for label_id in start_label_id_list]

        end_label_id_list = end_logit.argmax(axis=1).tolist()
        end_label_id_list = remove_pad_for_labels(feature.word_ids, end_label_id_list, feature.token_type_ids)
        end_label_list = [id_to_label_map[label_id] for label_id in end_label_id_list]

        entities = []

        previous_end_idx = None
        for start_idx, start_label in enumerate(start_label_list):
            if start_label == 'O':
                continue

            for end_idx, end_label in enumerate(end_label_list):
                if end_idx >= start_idx and start_label == end_label:
                    if  remove_overlap == 'short' and end_idx == previous_end_idx:
                        continue
                    elif  remove_overlap == 'long' and end_idx == previous_end_idx:
                        entities.pop()

                    entities.append(EntityExample(
                        start_idx=start_idx,
                        end_idx=end_idx+1,
                        entity=''.join(feature.words[start_idx: end_idx+1]),
                        type=start_label))
                    previous_end_idx = end_idx
                    break

        example_list.append(NerExample(text=feature.text, entities=entities, words=feature.words))
    return example_list



from typing import List
from siumaai.features.ner import NerExample


def calc_metric(gold_example_list: List[NerExample], pred_example_list: List[NerExample]):
    tp, fp, fn = 0, 0, 0
    for gold_example, pred_example in zip(gold_example_list, pred_example_list):
        _tp, _fp, _fn = 0, 0, 0
        for dst_entity in pred_example.entities:
            if dst_entity in gold_example.entities:
                _tp += 1
            else:
                _fp += 1
        _fn = len(gold_example.entities) - _tp
        tp += _tp
        fp += _fp
        fn += _fn
    precision = tp / (tp+fp) if tp+fp != 0 else 0
    recall = tp / (tp+fn) if tp+fn != 0 else 0
    f1 = 2 * precision * recall / (precision+recall) if precision+recall != 0 else 0
    return {
        'precision': precision,
        'recall': recall,
        'f1': f1,
        'tp': tp,
        'fp': fp,
        'fn': fn
    }


def calc_metric2(gold_example_list: List[NerExample], pred_example_list: List[NerExample]):
    tp_map, fp_map, fn_map = {}, {}, {}

    for gold_example, pred_example in zip(gold_example_list, pred_example_list):
        for pred_entity in pred_example.entities:
            if pred_entity in gold_example.entities:
                if pred_entity.type not in tp_map:
                    tp_map[pred_entity.type] = 0
                tp_map[pred_entity.type] += 1
            else:
                if pred_entity.type not in fp_map:
                    fp_map[pred_entity.type] = 0
                fp_map[pred_entity.type] += 1

        for gold_entity in gold_example.entities:
            if gold_entity not in pred_example.entities:
                if gold_entity.type not in fn_map:
                    fn_map[gold_entity.type] = 0
                fn_map[gold_entity.type] += 1

    label_set = set()
    label_set.update(tp_map.keys())
    label_set.update(fp_map.keys())
    label_set.update(fn_map.keys())

    metric = {
        label: {
            'precision': tp_map[label] / (tp_map[label]+fp_map[label]) if tp_map[label]+fp_map[label] != 0 else 0,
            'recall': tp_map[label] / (tp_map[label]+fn_map[label]) if tp_map[label]+fn_map[label] != 0 else 0,
            'f1': 2 * tp_map[label] / (2 * tp_map[label] + fn_map[label] + fp_map[label]) if 2 * tp_map[label] + fn_map[label] + fp_map[label] != 0 else 0
        }
        for label in label_set
    }

    metric['micro'] = {
        'precision': sum(tp_map.values()) / (sum(tp_map.values()) + sum(fp_map.values())) if sum(tp_map.values())+sum(fp_map.values()) != 0 else 0,
        'recall': sum(tp_map.values()) / (sum(tp_map.values()) + sum(fn_map.values())) if sum(tp_map.values())+sum(fn_map.values()) != 0 else 0,
        'f1': 2 * sum(tp_map.values()) / (2 * sum(tp_map.values()) + sum(fn_map.values()) + sum(fp_map.values())) if 2 * sum(tp_map.values()) + sum(fn_map.values()) + sum(fp_map.values()) != 0 else 0
    }

    metric['macro'] = {
        item: sum([metric[label][item] for label in label_set]) / len(label_set)
        for item in ['precision', 'recall', 'f1']
    }

    return metric



from .ner.crf_for_ner import CrfForNer
from .ner.span_for_ner import SpanForNer
from .ner.mrc_for_ner import MRCForNer
from .ner.global_pointer_for_ner import GlobalPointerForNer


MODEL_CLS_MAP = {
    'crf_for_ner': CrfForNer,
    'span_for_ner': SpanForNer,
    'mrc_for_ner': MRCForNer,
    'global_pointer_for_ner': GlobalPointerForNer
}



from torch import nn
from transformers import AutoModel 
from siumaai.layers.crf import CRF




class CrfForNer(nn.Module):
    def __init__(self, pretrain_model_path, num_labels, dropout_rate, hidden_size, vocab_len=None, pad_id=-100):
        super(CrfForNer, self).__init__()

        self.model = AutoModel.from_pretrained(pretrain_model_path)
        self.dropout = nn.Dropout(dropout_rate)
        self.classifier = nn.Linear(hidden_size, num_labels)
        self.crf = CRF(num_tags=num_labels, batch_first=True)
        self.pad_id = pad_id
        if vocab_len is not None:
            self.model.resize_token_embeddings(vocab_len)

    def forward(self, input_ids, token_type_ids=None, attention_mask=None, labels=None):
        outputs =self.model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)
        outputs = (logits,)
        if labels is not None:
            # mask = attention_mask & (labels != self.pad_id).long()
            # loss = self.crf(emissions=logits, tags=labels, mask=mask)
            loss = self.crf(emissions=logits, tags=labels, mask=attention_mask)
            outputs =(-1*loss,) + outputs
        else:
            crf_outputs = self.crf.decode(emissions=logits, mask=attention_mask)
            outputs = (crf_outputs, ) + outputs
        return outputs



import torch
from torch import nn


class FGM:
    def __init__(self, model: nn.Module, eps=1.0):
        if hasattr(model, 'module'):
            self.model = model.module
        else:
            self.model = model
        self.eps = eps
        self.backup = {}

    def attack(self, embedding_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and embedding_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm and not torch.isnan(norm):
                    param.data.add_(self.eps * param.grad / norm)

    def restore(self, embedding_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and embedding_name in name:
                assert name in self.backup, [name, self.backup.keys()]
                param.data = self.backup[name]
        self.backup = {}



import torch
from torch import nn

class PGD:
    def __init__(self, model: nn.Module, eps=1.0, alpha=0.3):
        if hasattr(model, 'module'):
            self.model = model.module
        else:
            self.model = model
        self.eps = eps
        self.alpha = alpha
        self.embedding_backup = {}
        self.grad_backup = {}

    def attack(self, embedding_name='word_embeddings', is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and embedding_name in name:
                if is_first_attack is True:
                    self.embedding_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)

                if norm and not torch.isnan(norm):
                    param.data.add_(self.alpha * param.grad / norm)
                    param.data = self.project(name, param.data)

    def restore(self, embedding_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and embedding_name in name:
                assert name in self.embedding_backup, [name, self.embedding_backup.keys()]
                param.data = self.embedding_backup[name]
        self.embedding_backup = {}

    def project(self, name, param_data):
        r = param_data - self.embedding_backup[name]
        if torch.norm(r) > self.eps:
            r = self.eps * r / torch.norm(r)
        return self.embedding_backup[name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                param.grad = self.grad_backup[name]
                 



import torch


def multilabel_categorical_crossentropy(y_pred, y_true):
    """
    https://kexue.fm/archives/7359
    """
    y_pred = (1 - 2 * y_true) * y_pred  # -1 -> pos classes, 1 -> neg classes
    y_pred_neg = y_pred - y_true * 1e12  # mask the pred outputs of pos classes
    y_pred_pos = ( y_pred - (1 - y_true) * 1e12 )  # mask the pred outputs of neg classes
    zeros = torch.zeros_like(y_pred[..., :1])
    y_pred_neg = torch.cat([y_pred_neg, zeros], dim=-1)
    y_pred_pos = torch.cat([y_pred_pos, zeros], dim=-1)
    neg_loss = torch.logsumexp(y_pred_neg, dim=-1)
    pos_loss = torch.logsumexp(y_pred_pos, dim=-1)

    return (neg_loss + pos_loss).mean()



import torch
import torch.nn as nn
import torch.nn.functional as F


class FocalLoss(nn.Module):
    def __init__(self, gamma=2, weight=None, ignore_index=-100, reduction='mean'):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.weight = weight
        self.ignore_index=ignore_index
        self.reduction = reduction

    def forward(self, input, target):
        """
        input: [N, C]
        target: [N, ]
        """
        logpt = F.log_softmax(input, dim=1)
        pt = torch.exp(logpt)
        logpt = (1 - pt) ** self.gamma * logpt
        return F.nll_loss(logpt, target, self.weight, ignore_index=self.ignore_index, reduction=self.reduction)



import torch
from typing import List, Optional, Union
from dataclasses import dataclass
from torch.utils.data import Dataset


@dataclass
class EntityExample:
    """
    [start_idx: end_idx) 注意尾部开区间
    """
    start_idx: int
    end_idx: int
    entity: str
    type: str


@dataclass
class NerExample:
    text: str
    entities: List[EntityExample]
    words: List[str]


@dataclass
class NerFeature:
    text: str
    entities: List[EntityExample]
    words: List[str]
    word_ids: List[Union[int, None]]
    input_ids: torch.Tensor
    attention_mask: torch.Tensor
    token_type_ids: torch.Tensor


# @dataclass
# class PairNerExample:
#     query: str
#     context: str
#     entities: List[EntityExample]
#     query_words: List[str]
#     context_words: List[str]
# 
# 
# @dataclass
# class PairNerFeature:
#     word_ids: List[Union[int, None]]
#     input_ids: torch.Tensor
#     attention_mask: torch.Tensor
#     token_type_ids: torch.Tensor


# class NerDataset(Dataset):
#     def __init__(self, example_list: List[NerExample]):
#         self.example_list = example_list
# 
#     def __len__(self):
#         return len(self.example_list)
# 
#     def __getitem__(self, idx) -> NerExample:
#         return self.example_list[idx]



import pytorch_lightning as pl
from layers.fgm import FGM, PGD


class FGM(pl.LightningModule):
    def __init__(self):
        super().__init__()
        self.automatic_optimization = False
        self.fgm = FGM(self.model)
        self.pgd = PGD(self.model)

    def training_step(self, batch, batch_idx):
        optimizer = self.optimizers()
        scheduler = self.lr_schedulers()
        assert hasattr(self, 'model')

        loss, *_ = self.model(**batch)
        optimizer.zero_grad()
        self.manual_backward(loss)

        self.fgm.attack()
        loss, *_ = self.model(**batch)
        self.manual_backward(loss)
        self.fgm.restore()

        self.trainer.train_loop.running_loss.append(loss)
        self.log('train_loss', loss, on_epoch=True)
        
        optimizer.step()
        scheduler.step()
        return loss

    def training_step(self, batch, batch_idx):
        K = 3
        optimizer = self.optimizers()
        scheduler = self.lr_schedulers()

        loss, *_ = self.model(**batch)
        optimizer.zero_grad()
        self.manual_backward(loss)
        self.pgd.backup_grad()

        for t in range(K):
            self.pgd.attack(is_first_attack=(t==0))
            if t != K-1:
                optimizer.zero_grad()
            else:
                self.pgd.restore_grad()

            loss, *_ = self.model(**batch)
            self.manual_backward(loss)
        self.pgd.restore()

        self.trainer.train_loop.running_loss.append(loss)
        self.log('train_loss', loss, on_epoch=True)
        
        optimizer.step()
        scheduler.step()
        return loss



from torch import nn
from transformers import AutoModel 
from siumaai.losses.focal_loss import FocalLoss
from siumaai.losses.label_smoothing import LabelSmoothingCrossEntropy




class SpanForNer(nn.Module):
    def __init__(self, pretrain_model_path, num_labels, dropout_rate, hidden_size, vocab_len=None, loss_type=None):
        super(SpanForNer, self).__init__()

        self.model = AutoModel.from_pretrained(pretrain_model_path)
        self.dropout = nn.Dropout(dropout_rate)
        self.start_fc = nn.Linear(hidden_size, num_labels)
        self.end_fc = nn.Linear(hidden_size, num_labels)
        self.num_labels = num_labels
        
        if loss_type == 'focal':
            self.criterion = FocalLoss()
        elif loss_type == 'ls_ce':
            self.criterion = LabelSmoothingCrossEntropy()
        else:
            self.criterion = nn.CrossEntropyLoss()

        if vocab_len is not None:
            self.model.resize_token_embeddings(vocab_len)

    def forward(self, input_ids, token_type_ids, attention_mask, start_labels=None, end_labels=None):
        outputs =self.model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)

        start_logits = self.start_fc(sequence_output)
        end_logits = self.end_fc(sequence_output)
        outputs = (start_logits, end_logits)

        if start_labels is not None and end_labels is not None:
            active_loss = attention_mask.view(-1) == 1

            active_start_logits = start_logits.view(-1, self.num_labels)[active_loss]
            active_end_logits = end_logits.view(-1, self.num_labels)[active_loss]

            active_start_labels = start_labels.view(-1)[active_loss]
            active_end_labels = end_labels.view(-1)[active_loss]

            start_loss = self.criterion(active_start_logits, active_start_labels)
            end_loss = self.criterion(active_end_logits, active_end_labels)
            outputs = (start_loss+end_loss, ) + outputs


        return outputs



import torch
from torch import nn
from transformers import AutoModel
from siumaai.losses.multilabel_categorical_crossentropy import multilabel_categorical_crossentropy


class GlobalPointerForNer(nn.Module):
    def __init__(self, pretrain_model_path, hidden_size, num_labels, inner_dim, RoPE=True, vocab_len=None):
        super().__init__()
        self.model = AutoModel.from_pretrained(pretrain_model_path)
        self.num_labels = num_labels
        self.inner_dim = inner_dim
        self.dense = nn.Linear(hidden_size, self.num_labels * self.inner_dim * 2)

        self.RoPE = RoPE

        if vocab_len is not None:
            self.model.resize_token_embeddings(vocab_len)
    
    
    def sinusoidal_position_embedding(self, batch_size, seq_len, output_dim):
        position_ids = torch.arange(0, seq_len, dtype=torch.float).unsqueeze(-1)

        indices = torch.arange(0, output_dim // 2, dtype=torch.float)
        indices = torch.pow(10000, -2 * indices / output_dim)
        embeddings = position_ids * indices
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = embeddings.repeat((batch_size, *([1]*len(embeddings.shape))))
        embeddings = torch.reshape(embeddings, (batch_size, seq_len, output_dim))
        embeddings = embeddings.to(self.device)
        return embeddings
        
    def forward(self, input_ids, attention_mask, token_type_ids, labels=None, criterion_mask=None):
        self.device = input_ids.device
        
        context_outputs = self.model(input_ids, attention_mask, token_type_ids)
        # last_hidden_state:(batch_size, seq_len, hidden_size)
        last_hidden_state = context_outputs[0]

        batch_size = last_hidden_state.size()[0]
        seq_len = last_hidden_state.size()[1]

        # outputs:(batch_size, seq_len, num_labels*inner_dim*2)
        outputs = self.dense(last_hidden_state)
        outputs = torch.split(outputs, self.inner_dim * 2, dim=-1)
        # outputs:(batch_size, seq_len, num_labels, inner_dim*2)
        outputs = torch.stack(outputs, dim=-2)
        # qw,kw:(batch_size, seq_len, num_labels, inner_dim)
        qw, kw = outputs[...,:self.inner_dim], outputs[...,self.inner_dim:] # TODO:修改为Linear获取？

        if self.RoPE:
            # pos_emb:(batch_size, seq_len, inner_dim)
            pos_emb = self.sinusoidal_position_embedding(batch_size, seq_len, self.inner_dim)
            # cos_pos,sin_pos: (batch_size, seq_len, 1, inner_dim)
            cos_pos = pos_emb[..., None, 1::2].repeat_interleave(2, dim=-1)
            sin_pos = pos_emb[..., None,::2].repeat_interleave(2, dim=-1)
            qw2 = torch.stack([-qw[..., 1::2], qw[...,::2]], -1)
            qw2 = qw2.reshape(qw.shape)
            qw = qw * cos_pos + qw2 * sin_pos
            kw2 = torch.stack([-kw[..., 1::2], kw[...,::2]], -1)
            kw2 = kw2.reshape(kw.shape)
            kw = kw * cos_pos + kw2 * sin_pos
            
        # logits:(batch_size, num_labels, seq_len, seq_len)
        logits = torch.einsum('bmhd,bnhd->bhmn', qw, kw)

        # padding mask
        pad_mask = attention_mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.num_labels, seq_len, seq_len)
        # pad_mask_h = attention_mask.unsqueeze(1).unsqueeze(-1).expand(batch_size, self.num_labels, seq_len, seq_len)
        # pad_mask = pad_mask_v&pad_mask_h
        logits = logits*pad_mask - (1-pad_mask)*1e12

        # 排除下三角
        mask = torch.tril(torch.ones_like(logits), -1) 
        logits = logits - mask * 1e12
        
        logits = logits/self.num_labels**0.5

        if labels is not None and criterion_mask is not None:

            criterion_mask = torch.triu(criterion_mask, 0).view(-1).bool()
            active_logits = logits.view(-1)[criterion_mask]
            active_labels = labels.view(-1)[criterion_mask]
            loss = multilabel_categorical_crossentropy(active_logits, active_labels)

            # loss = multilabel_categorical_crossentropy(
            #         logits.view(batch_size*self.num_labels, -1), 
            #         labels.view(batch_size*self.num_labels, -1))
            return loss, logits
        else:
            return (logits,)



        
        # return logits/self.num_labels**0.5



import torch
from dataclasses import dataclass
from torch.utils.data import Dataset
from typing import Dict, Optional, List, Any, Union
from . import NerExample, NerFeature, EntityExample
from ..utils import add_pad_for_labels, remove_pad_for_labels, compare_words_tokens, add_pad_for_2d_labels, remove_pad_for_2d_labels


@dataclass
class MRCForNerExample:
    query_words: List
    context_words: List
    text: str
    query_type: Any
    start_labels: List
    end_labels: List
    span_labels: List[List]
    example_id: int


@dataclass
class MRCForNerFeature:
    query_type: Any
    example_id: int
    query_words: List[str]
    context_words: List[str]
    word_ids: List[Union[int, None]]
    input_ids: torch.Tensor
    attention_mask: torch.Tensor
    token_type_ids: torch.Tensor
    start_labels: torch.Tensor
    end_labels: torch.Tensor
    span_labels: torch.Tensor
    start_criterion_mask: torch.Tensor
    end_criterion_mask: torch.Tensor
    span_criterion_mask: torch.Tensor



class MRCForNerDataset(Dataset):
    def __init__(self, 
            example_list: List[NerExample], 
            tokenizer, 
            label_to_query_map,
            max_seq_length, 
            pad_id=-100, 
            label_all_tokens=False, 
            check_tokenization=False,
            lazy_load=False):
        self.example_list = example_list
        self.tokenizer = tokenizer
        self.label_to_query_map = label_to_query_map
        self.max_seq_length = max_seq_length
        self.pad_id = pad_id
        self.label_all_tokens = label_all_tokens
        self.check_tokenization = check_tokenization
        self.lazy_load = lazy_load

        self.mrc_example_list = []
        for example_id, example in enumerate(self.example_list):
            for query_type, query in label_to_query_map.items():
            
                start_labels = [0. for _ in example.words]
                end_labels = [0. for _ in example.words]
                span_labels = [[0. for _ in example.words] for _ in example.words]

                for entity in example.entities:
                    if entity.type != query_type:
                        continue
                    if len(example.words) < entity.end_idx or ''.join(example.words[entity.start_idx: entity.end_idx]) != entity.entity:
                        guess_start = example.text.index(entity.entity)
                        guess_end = guess_start + len(entity.entity)
                        raise Exception(f'index error, text: {example.text}, item: {entity}, max_len: {len(example.words)}, guess: {(guess_start, guess_end)}')

                    start_labels[entity.start_idx] = 1.0
                    end_labels[entity.end_idx-1] = 1.0
                    span_labels[entity.start_idx][entity.end_idx-1] = 1.0
                self.mrc_example_list.append(MRCForNerExample(
                    list(query),
                    example.words,
                    example.text,
                    query_type,
                    start_labels,
                    end_labels,
                    span_labels,
                    example_id
                    ))
        

        if self.lazy_load is False:
            self.feature_list = convert_examples_to_feature(
                    self.tokenizer, 
                    self.mrc_example_list, 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)

    def __len__(self):
        # return len(self.feature_list)
        return len(self.mrc_example_list)

    def __getitem__(self, idx) -> MRCForNerFeature:
        if self.lazy_load is False:
            return self.feature_list[idx]
        else:
            return convert_examples_to_feature(
                    self.tokenizer, 
                    [self.mrc_example_list[idx]], 
                    self.max_seq_length, 
                    self.pad_id,
                    label_all_tokens=self.label_all_tokens,
                    check_tokenization=self.check_tokenization)[0]

def convert_examples_to_feature(
        tokenizer,
        example_list: List[MRCForNerExample],
        max_seq_length: int,
        pad_id: int=-100,
        label_all_tokens: bool=False,
        check_tokenization=False
        ) -> List[MRCForNerFeature]:

    
    encoded_inputs = tokenizer(
        [example.query_words for example in example_list],
        [example.context_words for example in example_list],
        is_split_into_words=True,
        truncation=True,
        padding='max_length',
        max_length=max_seq_length
    )

    feature_list = []
    for index, example in enumerate(example_list):
        word_ids = encoded_inputs.word_ids(index)
        assert check_tokenization is False or compare_words_tokens(tokenizer.convert_ids_to_tokens(encoded_inputs['input_ids'][index]), example.query_words+example.context_words, word_ids), example
        token_type_ids = encoded_inputs['token_type_ids'][index]

        # fix labels
        start_labels, start_criterion_mask = add_pad_for_labels(
                word_ids, 
                example.start_labels, 
                pad_id, 
                token_type_ids, 
                label_all_tokens=label_all_tokens, 
                required_token_type_id=1,
                return_criterion_mask=True)
        end_labels, end_criterion_mask = add_pad_for_labels(
                word_ids, 
                example.end_labels, 
                pad_id, 
                token_type_ids, 
                label_all_tokens=label_all_tokens, 
                required_token_type_id=1,
                return_criterion_mask=True)
        span_labels, span_criterion_mask = add_pad_for_2d_labels(
                word_ids, 
                example.span_labels, 
                pad_id, 
                token_type_ids, 
                label_all_tokens=label_all_tokens, 
                required_token_type_id=1,
                return_criterion_mask=True)


        feature_list.append(MRCForNerFeature(
            query_type=example.query_type,
            example_id=example.example_id,
            query_words=example.query_words,
            context_words=example.context_words,
            word_ids=word_ids,
            input_ids=torch.tensor(encoded_inputs['input_ids'][index]),
            attention_mask=torch.tensor(encoded_inputs['attention_mask'][index]),
            token_type_ids=torch.tensor(encoded_inputs['token_type_ids'][index]),
            start_labels=torch.tensor(start_labels),
            end_labels=torch.tensor(end_labels),
            span_labels=torch.tensor(span_labels),
            start_criterion_mask=torch.tensor(start_criterion_mask),
            end_criterion_mask=torch.tensor(end_criterion_mask),
            span_criterion_mask=torch.tensor(span_criterion_mask)
        ))

    return feature_list


def convert_logits_to_examples(
        feature_list: List[MRCForNerFeature], 
        start_logits: torch.Tensor, 
        end_logits: torch.Tensor, 
        span_logits: torch.Tensor,
        remove_overlap=None) -> List[NerExample]:

    example_list = []
    group_example_id_map = {}
    group_example_list = []

    for index, (start_logit, end_logit, span_logit, feature) in enumerate(zip(start_logits, end_logits, span_logits, feature_list)):
        example_id = feature.example_id
        query_type = feature.query_type

        start_labels = torch.where(start_logit>0, 1, 0).tolist()
        start_labels = remove_pad_for_labels(
                feature.word_ids, start_labels, feature.token_type_ids, required_token_type_id=1)

        end_labels = torch.where(end_logit>0, 1, 0).tolist()
        end_labels = remove_pad_for_labels(
                feature.word_ids, end_labels, feature.token_type_ids, required_token_type_id=1)

        span_labels = torch.where(span_logit>0, 1, 0).tolist()
        span_labels = remove_pad_for_2d_labels(
                feature.word_ids, span_labels, feature.token_type_ids, required_token_type_id=1)

        entities = []
        previous_end_index = None
        for start_index, start_label in enumerate(start_labels):
            if start_label != 1:
                continue

            for end_index, end_label in enumerate(end_labels):
                if end_index >= start_index and start_label == end_label == span_labels[start_index][end_index]:
                    if  remove_overlap == 'short' and end_index == previous_end_index:
                        continue
                    elif  remove_overlap == 'long' and end_index == previous_end_index:
                        entities.pop()

                    entities.append(EntityExample(
                        start_idx=start_index,
                        end_idx=end_index+1,
                        entity=''.join(feature.context_words[start_index: end_index+1]),
                        type=query_type))
                    previous_end_index = end_index
                    break

        example_list.append(NerExample(
            text=''.join(feature.context_words), entities=entities, words=feature.context_words))
        if example_id not in group_example_id_map:
            group_example_id_map[example_id] = []
        group_example_id_map[example_id].append(index)

    for _, group_index_list in sorted(group_example_id_map.items(), key=lambda item: item[0]):
        group_entities = []
        for group_index in group_index_list:
            group_entities.extend(example_list[group_index].entities)
        group_example_list.append(NerExample(
                text=example_list[group_index_list[0]].text,
                words=example_list[group_index_list[0]].text,
                entities=sorted(group_entities, key=lambda entity: entity.start_idx)))


    return group_example_list




from setuptools import setup, find_packages


setup(
    name='siumaai',
    version='0.0.2',
    description='a siumaai for nlp',
    license='Apache License 2.0',
    author='Zonzely',
    install_requires=[
        'torch',
        'transformers',
        'pytorch-lightning'
    ],
    packages=find_packages()
)



import pytorch_lightning as pl
from functools import lru_cache
from transformers import AdamW, get_linear_schedule_with_warmup


class Ner(pl.LightningModule):
    def __init__(self, 
            learning_rate, 
            adam_epsilon, 
            warmup_rate, 
            weight_decay,
            model_cls,
            **model_kwargs):
        super().__init__()
        self.save_hyperparameters()
        self.model = model_cls(**model_kwargs)


    # @lru_cache()
    # def total_steps(self):
    #     import ipdb; ipdb.set_trace()
    #     return len(self.train_dataloader()) // self.trainer.accumulate_grad_batches * self.trainer.max_epochs

    def configure_optimizers(self):

        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p 
                    for n, p in self.model.named_parameters() 
                    if not any(nd in n for nd in no_decay)],
                "weight_decay": self.hparams.weight_decay,
            },
            {
                "params": [
                    p 
                    for n, p in self.model.named_parameters() 
                    if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]

        optimizer = AdamW(
                optimizer_grouped_parameters, 
                lr=self.hparams.learning_rate, 
                eps=self.hparams.adam_epsilon)
        # num_train_steps = self.total_steps()
        # num_warmup_steps = int(self.hparams.warmup_rate * num_train_steps)
        scheduler = get_linear_schedule_with_warmup(
                optimizer, 
                num_warmup_steps=int(self.hparams.warmup_rate * self.trainer.estimated_stepping_batches), 
                num_training_steps=self.trainer.estimated_stepping_batches)
        return {
            'optimizer': optimizer,
            'lr_scheduler': {
                'scheduler': scheduler,
                'interval': 'step',
                'frequency': 1
            }
        }
    

    def forward(self, input_ids, token_type_ids, attention_mask):
        return self.model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)

    def training_step(self, batch, batch_idx):

        loss, *_ = self.model(**batch)
        self.log('train_loss', loss, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):

        loss, *_ = self.model(**batch)
        self.log('val_loss', loss, on_epoch=True, prog_bar=True)



class CrfNer(Ner):
    def __init__(self, 
            crf_learning_rate,
            learning_rate, 
            adam_epsilon, 
            warmup_rate, 
            weight_decay,
            model_cls,
            **model_kwargs):
        super().__init__(learning_rate, adam_epsilon, warmup_rate, weight_decay, model_cls, **model_kwargs)
        self.save_hyperparameters()

    def configure_optimizers(self):

        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p 
                    for n, p in self.model.named_parameters() 
                    if 'crf' in n],
                "weight_decay": self.hparams.weight_decay,
                "lr": self.hparams.crf_learning_rate
            },
            {
                "params": [
                    p 
                    for n, p in self.model.named_parameters() 
                    if not any(nd in n for nd in no_decay) and 'crf' not in n],
                "weight_decay": self.hparams.weight_decay,
            },
            {
                "params": [
                    p 
                    for n, p in self.model.named_parameters() 
                    if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            }
        ]

        optimizer = AdamW(
                optimizer_grouped_parameters, 
                lr=self.hparams.learning_rate, 
                eps=self.hparams.adam_epsilon)
        # num_train_steps = self.total_steps()
        # num_warmup_steps = int(self.hparams.warmup_rate * num_train_steps)
        scheduler = get_linear_schedule_with_warmup(
                optimizer, 
                num_warmup_steps=int(self.hparams.warmup_rate * self.trainer.estimated_stepping_batches), 
                num_training_steps=self.trainer.estimated_stepping_batches)
        return {
            'optimizer': optimizer,
            'lr_scheduler': {
                'scheduler': scheduler,
                'interval': 'step',
                'frequency': 1
            }
        }
